from typing import List, Tuple

from django.db import models
from django.utils.timezone import now

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveContactQuerySet(models.QuerySet):
    def grouped_by_standing(self) -> dict:
        """returns alliance contacts grouped by their standing as dict"""

        contacts_by_standing = dict()
        for contact in self.all():
            if contact.standing not in contacts_by_standing:
                contacts_by_standing[contact.standing] = set()
            contacts_by_standing[contact.standing].add(contact)

        return contacts_by_standing


class EveContactManager(models.Manager):
    def get_queryset(self) -> models.QuerySet:
        return EveContactQuerySet(self.model, using=self._db)


class EveEntityManager(models.Manager):
    def create_from_esi_contact(
        self, contact_id: int, contact_type: str
    ) -> models.Model:
        return self.create(
            id=contact_id, category=self.model.Category.from_esi_type(contact_type)
        )

    def get_or_create_from_esi_contact(
        self, contact_id: int, contact_type: str
    ) -> models.Model:
        return self.get_or_create(
            id=contact_id,
            defaults={"category": self.model.Category.from_esi_type(contact_type)},
        )

    def get_or_create_from_esi_info(self, info) -> Tuple[models.Model, bool]:
        """returns an EveEntity object for the given esi info
        will return existing or create new one if needed
        """
        id = info.get("alliance_id") or info.get("corporation_id")
        category = (
            self.model.Category.ALLIANCE
            if info.get("alliance_id")
            else self.model.Category.CORPORATION
        )
        return self.get_or_create(id=id, defaults={"category": category})


class EveWarManager(models.Manager):
    def active_wars(self) -> models.QuerySet:
        return self.filter(started__lt=now(), finished__gt=now()) | self.filter(
            started__lt=now(), finished__isnull=True
        )

    def finished_wars(self) -> models.QuerySet:
        return self.filter(finished__lte=now())

    def war_targets(self, alliance_id: int) -> List[models.Model]:
        """returns list of current war targets for given alliance as EveEntity objects
        or an empty list if there are none
        """
        war_targets = list()
        active_wars = self.active_wars()
        # case 1 alliance is aggressor
        for war in active_wars:
            if war.aggressor_id == alliance_id:
                war_targets.append(war.defender)
                if war.allies:
                    war_targets += list(war.allies.all())

        # case 2 alliance is defender
        for war in active_wars:
            if war.defender_id == alliance_id:
                war_targets.append(war.aggressor)

        # case 3 alliance is ally
        for war in active_wars:
            if war.allies.filter(id=alliance_id).exists():
                war_targets.append(war.aggressor)

        return war_targets

    def update_from_esi(self, id: int):
        from .models import EveEntity

        logger.info("Retrieving war details for ID %s", id)
        war_info = esi.client.Wars.get_wars_war_id(war_id=id).results()
        finished = war_info.get("finished")
        if finished and finished <= now():
            logger.info("Ignoring finished war with ID %s", id)
            return

        logger.info("Updating war details for ID %s", id)
        try:
            war = self.get(id=id)
        except self.model.DoesNotExist:
            aggressor = EveEntity.objects.get_or_create_from_esi_info(
                war_info.get("aggressor")
            )[0]
            defender = EveEntity.objects.get_or_create_from_esi_info(
                war_info.get("defender")
            )[0]
            war = self.create(
                id=id,
                aggressor=aggressor,
                declared=war_info.get("declared"),
                defender=defender,
                is_mutual=war_info.get("mutual"),
                is_open_for_allies=war_info.get("open_for_allies"),
                retracted=war_info.get("retracted"),
                started=war_info.get("started"),
                finished=war_info.get("finished"),
            )

        else:
            self.update(
                retracted=war_info.get("retracted"),
                started=war_info.get("started"),
                finished=war_info.get("finished"),
                is_mutual=war_info.get("mutual"),
                is_open_for_allies=war_info.get("open_for_allies"),
            )
            war.allies.clear()

        if war_info.get("allies"):
            for ally_info in war_info.get("allies"):
                eve_entity = EveEntity.objects.get_or_create_from_esi_info(ally_info)[0]
                war.allies.add(eve_entity)



from unittest.mock import patch

from django.test import TestCase

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import NoSocketsTestCase, generate_invalid_pk

from .. import tasks
from ..models import EveContact, EveEntity, SyncedCharacter, SyncManager
from . import (
    ALLIANCE_CONTACTS,
    BravadoOperationStub,
    LoadTestDataMixin,
    create_test_user,
)

TASKS_PATH = "standingssync.tasks"
MODELS_PATH = "standingssync.models"


@patch(TASKS_PATH + ".run_manager_sync")
@patch(TASKS_PATH + ".update_all_wars")
class TestRunRegularSync(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # given
        cls.user_1 = create_test_user(cls.character_1)
        cls.main_ownership_1 = CharacterOwnership.objects.get(
            character=cls.character_1, user=cls.user_1
        )

    def test_should_start_all_tasks(self, mock_update_all_wars, mock_run_manager_sync):
        # given
        sync_manager = SyncManager.objects.create(
            alliance=self.alliance_1,
            character_ownership=self.main_ownership_1,
            version_hash="new",
        )
        with patch(TASKS_PATH + ".is_esi_online", lambda: True):
            # when
            tasks.run_regular_sync()
        # then
        self.assertTrue(mock_update_all_wars.delay.called)
        args, _ = mock_run_manager_sync.delay.call_args
        self.assertEqual(args[0], sync_manager.pk)

    def test_abort_when_esi_if_offline(
        self, mock_update_all_wars, mock_run_manager_sync
    ):
        # given
        with patch(TASKS_PATH + ".is_esi_online", lambda: False):
            # when
            tasks.run_regular_sync()
        # then
        self.assertFalse(mock_update_all_wars.delay.called)
        self.assertFalse(mock_run_manager_sync.delay.called)


class TestCharacterSync(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # 1 user with 1 alt character
        cls.user_1 = create_test_user(cls.character_1)
        cls.main_ownership_1 = CharacterOwnership.objects.get(
            character=cls.character_1, user=cls.user_1
        )
        alt_ownership_2 = CharacterOwnership.objects.create(
            character=cls.character_2, owner_hash="x2", user=cls.user_1
        )
        alt_ownership_3 = CharacterOwnership.objects.create(
            character=cls.character_3, owner_hash="x3", user=cls.user_1
        )

        # sync manager with contacts
        cls.sync_manager = SyncManager.objects.create(
            alliance=cls.alliance_1,
            character_ownership=cls.main_ownership_1,
            version_hash="new",
        )
        for contact in ALLIANCE_CONTACTS:
            EveContact.objects.create(
                manager=cls.sync_manager,
                eve_entity=EveEntity.objects.get(id=contact["contact_id"]),
                standing=contact["standing"],
                is_war_target=False,
            )

        # sync char
        cls.synced_character_2 = SyncedCharacter.objects.create(
            character_ownership=alt_ownership_2, manager=cls.sync_manager
        )
        cls.synced_character_3 = SyncedCharacter.objects.create(
            character_ownership=alt_ownership_3, manager=cls.sync_manager
        )

    def test_run_character_sync_wrong_pk(self):
        """calling for an non existing sync character should raise an exception"""
        with self.assertRaises(SyncedCharacter.DoesNotExist):
            tasks.run_character_sync(generate_invalid_pk(SyncedCharacter))

    @patch(TASKS_PATH + ".SyncedCharacter.update")
    def test_should_call_update(self, mock_update):
        # given
        mock_update.return_value = True
        # when
        result = tasks.run_character_sync(self.synced_character_2)
        # then
        self.assertTrue(result)
        self.assertTrue(mock_update.called)


@patch(TASKS_PATH + ".run_character_sync")
class TestManagerSync(LoadTestDataMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # create environment
        # 1 user has permission for manager sync
        cls.user_1 = create_test_user(cls.character_1)
        cls.main_ownership_1 = CharacterOwnership.objects.get(
            character=cls.character_1, user=cls.user_1
        )
        cls.user_1 = AuthUtils.add_permission_to_user_by_name(
            "standingssync.add_syncmanager", cls.user_1
        )

        # user 1 has no permission for manager sync and has 1 alt
        cls.user_2 = create_test_user(cls.character_2)
        cls.main_ownership_2 = CharacterOwnership.objects.get(
            character=cls.character_2, user=cls.user_2
        )
        cls.alt_ownership_2 = CharacterOwnership.objects.create(
            character=cls.character_4, owner_hash="x4", user=cls.user_2
        )

    # run for non existing sync manager
    def test_run_sync_wrong_pk(self, mock_run_character_sync):
        with self.assertRaises(SyncManager.DoesNotExist):
            tasks.run_manager_sync(99999)

    @patch(MODELS_PATH + ".SyncManager.update_from_esi")
    def test_should_report_error_when_unexpected_exception_occurs(
        self, mock_update_from_esi, mock_run_character_sync
    ):
        # given
        mock_update_from_esi.side_effect = RuntimeError
        sync_manager = SyncManager.objects.create(
            alliance=self.alliance_1, character_ownership=self.main_ownership_1
        )
        # when
        result = tasks.run_manager_sync(sync_manager.pk)
        # then
        sync_manager.refresh_from_db()
        self.assertFalse(result)
        self.assertEqual(sync_manager.last_error, SyncManager.Error.UNKNOWN)

    @patch(MODELS_PATH + ".SyncManager.update_from_esi")
    def test_should_normally_run_character_sync(
        self, mock_update_from_esi, mock_run_character_sync
    ):
        # given
        mock_update_from_esi.return_value = "abc"
        sync_manager = SyncManager.objects.create(
            alliance=self.alliance_1, character_ownership=self.main_ownership_1
        )
        synced_character = SyncedCharacter.objects.create(
            character_ownership=self.alt_ownership_2, manager=sync_manager
        )
        # when
        result = tasks.run_manager_sync(sync_manager.pk)
        # then
        sync_manager.refresh_from_db()
        self.assertTrue(result)
        self.assertEqual(sync_manager.last_error, SyncManager.Error.NONE)
        args, kwargs = mock_run_character_sync.delay.call_args
        self.assertEqual(kwargs["sync_char_pk"], synced_character.pk)
        self.assertFalse(kwargs["force_sync"])


class TestUpdateWars(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @patch(TASKS_PATH + ".update_war")
    @patch(TASKS_PATH + ".esi")
    def test_should_start_tasks_for_each_war_id(self, mock_esi, mock_update_war):
        # given
        mock_esi.client.Wars.get_wars.return_value = BravadoOperationStub([1, 2, 3])
        # when
        tasks.update_all_wars()
        # then
        result = {row[0][0] for row in mock_update_war.delay.call_args_list}
        self.assertSetEqual(result, {1, 2, 3})

    @patch(TASKS_PATH + ".EveWar.objects.update_from_esi")
    def test_should_update_war(self, mock_update_from_esi):
        # when
        tasks.update_war(42)
        # then
        args, _ = mock_update_from_esi.call_args
        self.assertEqual(args[0], 42)



from django.contrib import admin

from . import tasks
from .models import SyncedCharacter, SyncManager


@admin.register(SyncedCharacter)
class SyncedCharacterAdmin(admin.ModelAdmin):
    list_display = (
        "user",
        "character_name",
        "version_hash",
        "_sync_ok",
        "last_sync",
        "last_error",
        "manager",
    )
    list_filter = (
        "last_error",
        "version_hash",
        "last_sync",
        "character_ownership__user",
        "manager",
    )
    actions = ["start_sync_contacts"]
    list_display_links = None

    @admin.display(boolean=True)
    def _sync_ok(self, obj) -> bool:
        return obj.is_sync_ok

    def has_add_permission(self, request):
        return False

    def user(self, obj):
        return obj.character_ownership.user

    def character_name(self, obj):
        return obj.__str__()

    @admin.display(description="Sync selected characters")
    def start_sync_contacts(self, request, queryset):
        names = list()
        for obj in queryset:
            tasks.run_character_sync.delay(sync_char_pk=obj.pk, force_sync=True)
            names.append(str(obj))
        self.message_user(request, "Started syncing for: {}".format(", ".join(names)))


@admin.register(SyncManager)
class SyncManagerAdmin(admin.ModelAdmin):
    list_display = (
        "alliance_name",
        "contacts_count",
        "synced_characters_count",
        "user",
        "character_name",
        "version_hash",
        "_sync_ok",
        "last_sync",
        "last_error",
    )
    list_display_links = None
    actions = ["start_sync_managers"]

    @admin.display(boolean=True)
    def _sync_ok(self, obj) -> bool:
        return obj.is_sync_ok

    def has_add_permission(self, request):
        return False

    def user(self, obj):
        return obj.character_ownership.user if obj.character_ownership else None

    def character_name(self, obj):
        return obj.__str__()

    def alliance_name(self, obj):
        return obj.alliance.alliance_name

    def contacts_count(self, obj):
        return "{:,}".format(obj.contacts.count())

    def synced_characters_count(self, obj):
        return "{:,}".format(obj.synced_characters.count())

    @admin.display(description="Sync selected managers")
    def start_sync_managers(self, request, queryset):
        names = list()
        for obj in queryset:
            tasks.run_manager_sync.delay(manager_pk=obj.pk, force_sync=True)
            names.append(str(obj))
        text = "Started syncing for: {} ".format(", ".join(names))
        self.message_user(request, text)
